"""
(Deprecated and unused)
This module offers basic connections between one client and one server.

Jiangchen Zhu  <zjcsjtu@umich.edu>
"""

import json
import socket


def sock_init(port_num):
    """Initialize a scoket that listens on port_num."""
    server_socket = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM)
    server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server_socket.bind(('localhost', port_num))
    server_socket.listen(10)
    # cancel the timeout of 3 seconds
    # server_socket.settimeout(3)
    return server_socket


def recv_data(ssocket):
    """Note that this only allows one end to send any data."""
    # receive a header which contains the size of msg
    # print('Receiving data header...')
    try:
        csock, dummy = ssocket.accept()
        header = csock.recv(1024).decode()
    except socket.timeout:
        print('Error: failed to receive data!')
    size = json.loads(header)['size']
    csock.close()
    # print('size =', size)

    # receive the actual data
    # print('Receiving some data...')
    try:
        csock, dummy = ssocket.accept()
        data = csock.recv(size).decode()
    except socket.timeout:
        print('Error: failed to receive data!')
    data = json.loads(data)
    csock.close()
    # print(data['data'])

    return data['data']


def send_data(target_port, data):
    """Send data to the target port."""
    msg = json.dumps({'data':data})
    sz = len(msg)
    # send a header indicating the size
    try:
        temp_socket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        temp_socket.connect(("localhost", target_port))
        header = {'size': sz}
        header_msg = json.dumps(header)
        temp_socket.send(header_msg.encode())
        temp_socket.close()
    except socket.error:
        print('Error: failed to send header!')

    # send the actual data to the target port
    try:
        temp_socket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        temp_socket.connect(("localhost", target_port))
        temp_socket.send(msg.encode())
        temp_socket.close()
    except socket.error:
        print('Error: failed to send data!')



"""
A simple implementation of a suffix tree, as specified in the paper,
used in Substring-Searchable Encryption.

Jiangchen Zhu  <zjcsjtu@umich.edu>
"""
from substr_enc.utils.suffix_tree_helper import get_path_helper, get_common_prefix
from substr_enc.utils.suffix_tree_helper import get_initial_path_helper
from substr_enc.utils.suffix_tree_helper import get_ind_helper

class Node:
    """Represent a node in suffix tree."""
    def __init__(self):
        """Initialize the node."""
        self.edge_label = ''
        self.parent = None
        self.children = []
        # if this is a leaf node, leaf_label represents the starting index
        self.leaf_label = -1


    def get_path(self):
        """Call the helper function and return the path label (incl. $)."""
        return get_path_helper(self)


    def get_initial_path(self):
        """Call the helper function and return the initial path label."""
        return get_initial_path_helper(self)


    def get_len(self):
        """Return he length of inital path (may incl. $)."""
        return len(self.get_initial_path())


    def get_ind(self):
        """Return the index in string of the first occurrence of path label."""
        node_inds = []
        get_ind_helper(self, node_inds)
        return min(node_inds)


def init_helper(cur_node, suffix, index):
    """Helper function for init method, recursively go down the tree."""
    if suffix == '':
        return
    target_child = None
    # check if there exists an edge starts with the same prefix
    for child in cur_node.children:
        prefix_tmp = get_common_prefix(suffix, child.edge_label)
        if prefix_tmp:
            target_child = child
            break

    # if we cannot find a child starts with the same prefix
    if target_child == None:
        target_child = Node()
        # the cur_node is no longer a leaf node
        cur_node.leaf_label = -1
        cur_node.children.append(target_child)
        # make a new child for the suffix
        target_child.parent = cur_node
        target_child.edge_label = suffix
        target_child.leaf_label = index
        return
    else:
        common_prefix = get_common_prefix(suffix, target_child.edge_label)
        common_pre_len = len(common_prefix)
        if common_pre_len < len(target_child.edge_label):
            # if the length of common prefix is shorter than the edge of
            # target child, then,
            # need to split the target_child into two nodes
            new_node = Node()
            new_node.leaf_label = target_child.leaf_label
            new_node.children = target_child.children.copy()
            new_node.edge_label = target_child.edge_label[common_pre_len:]
            target_child.leaf_label = -1
            target_child.children.clear()
            target_child.children.append(new_node)
            target_child.edge_label = common_prefix
            new_node.parent = target_child
            init_helper(target_child, suffix[common_pre_len:], index)
        else:
            # recursively build the tree for suffix
            init_helper(target_child, suffix[common_pre_len:], index)


def get_leaves_helper(node, output):
    """Preorder traverse the tree and get the leaf nodes rooted at this node."""
    if node == None:
        return
    if not node.children:
        output.append(node)
    for child in node.children:
        get_leaves_helper(child, output)


class SuffixTree:
    """Implement a suffix tree according to the paper."""
    def __init__(self, s):
        """Build a suffix tree given string s."""
        # insert dollar sign to make sure that a suffix tree exists
        # assume that $ does not appear in string s
        s += '$'
        self.root = Node()
        self.leaves = []
        # create nodes for every suffix
        for i in range(2, len(s) + 1):
            sufx = s[-i:]
            init_helper(cur_node=self.root, suffix=sufx, index=len(s) - i)
        get_leaves_helper(self.root, self.leaves)


    def get_leafpos(self, node):
        """
        Return the position (between 0 and n - 1) in the
        tree of the leftmost leaf in the subtree rooted at node.
        """
        # get all leaves that are rooted at node
        leaves_rooted_node = []
        get_leaves_helper(node, leaves_rooted_node)
        min_leafpos = len(self.leaves)
        for leaf in leaves_rooted_node:
            leaf_ind = self.leaves.index(leaf)
            if leaf_ind < min_leafpos:
                min_leafpos = leaf_ind
        return min_leafpos


    def get_num(self, node):
        """Return number of leaves in the subtree rooted at node."""
        leaves_rooted_node = []
        get_leaves_helper(node, leaves_rooted_node)
        return len(leaves_rooted_node)


    def show_tree(self):
        """Print the suffix tree level by level for debug."""
        print('Printing the suffix tree level by level...')
        tmp_list = list(self.root.children)
        ind = 1
        while tmp_list:
            print('level', ind)
            next_children = []
            for child in tmp_list:
                print(child.edge_label, '\t', child.leaf_label)
                next_children = next_children + list(child.children)
            tmp_list = next_children
            ind += 1
        print('------------------------------------------------')


    def get_nodes(self, node, output):
        """Preorder traverse the tree and get nodes."""
        if node == None:
            return
        output.append(node)
        for child in node.children:
            self.get_nodes(child, output)



"""
Implementation of Enc function in the paper.

Jiangchen Zhu  <zjcsjtu@umich.edu>
Enhao Zhang    <ehzhang@umich.edu>
Zhiqi Chen     <zhiqich@umich.edu>
"""
from substr_enc.utils import Node, SuffixTree
from substr_enc.model import LAMBDA
import random
import secrets
import hashlib
from Cryptodome.Cipher import AES
from Cryptodome import Random
import operator
import hmac

def key_gen():
    """Generate the keys according to paper."""
    key_list = []
    for i in range(7):
        h = hashlib.blake2b(key = secrets.token_bytes(LAMBDA), digest_size = LAMBDA)
        key_list.append(h.digest())
    return key_list


def encrypt(k, s):
    """
    Implement an Enc function in the paper.

    k: the key list generated by key_gen
    s: the original texts
    return: ciphertext tuple (d, c, l)
    """
    # build a tree upon the string s
    tree = SuffixTree(s)
    kd = k[0]
    kc = k[1]
    kl = k[2]
    k1 = k[3]
    k2 = k[4]
    k3 = k[5]
    k4 = k[6]
    d = {}
    nodes = []
    tree.get_nodes(tree.root, nodes)
    nodes.sort(key = operator.attrgetter('edge_label'))
    for node in nodes:
        children = []
        g2 = []
        for child in node.children:
            children.append(child)
            h2 = hmac.new(key = k2, digestmod = hashlib.sha256)
            h2.update(child.get_initial_path().encode('utf-8'))
            g2.append(h2.digest())
        for i in range(len(children), 128):
            h = hmac.new(key = secrets.token_bytes(LAMBDA), digestmod = hashlib.sha256)
            g2.append(h.digest())
        piu = [i for i in range(0, 128)]
        random.Random(hashlib.blake2b(key = secrets.token_bytes(LAMBDA), digest_size = LAMBDA)).shuffle(piu)
        f2 = [i for i in range(0, 128)]
        for i in range(0, 128):
            f2[i] = g2[piu[i]]
        h1 = hmac.new(key = k1, digestmod = hashlib.sha256)
        h1.update(node.get_initial_path().encode('utf-8'))
        f1 = h1.digest()
        xu = str(node.get_ind()) + '$' + str(tree.get_leafpos(node)) + '$' + str(tree.get_num(node)) + '$' + str(node.get_len()) + '$' + f1.hex()
        for i in range(0, 128):
            xu += '$' + f2[i].hex()
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(kd, AES.MODE_CFB, iv)
        wu = iv.hex() + '$' + cipher.encrypt(xu.encode("utf8")).hex()
        vu = ''
        for i in range(0, 128):
            vu += f2[i].hex() + '$'
        vu += wu
        d[f1] = vu
    for i in range(0, 2 * len(s) - len(nodes)):
        dummy_string = []
        for j in range(0, 129):
            h = hashlib.blake2b(key = secrets.token_bytes(LAMBDA), digest_size = LAMBDA)
            dummy_string.append(h.digest())
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(kd, AES.MODE_CFB, iv)
        enc0 = iv.hex() + '$' + cipher.encrypt(str(0).encode("utf8")).hex()
        dummy = ''
        for i in range(0, 128):
            dummy += dummy_string[i].hex() + '$'
        dummy += enc0
        d[dummy_string[128]] = dummy
    c = [i for i in range(0, len(s))]
    p = [i for i in range(0, len(s))]
    random.seed(k3)
    random.shuffle(p)
    for i in range(0, len(s)):
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(kc, AES.MODE_CFB, iv)
        c[p[i]] = iv.hex() + '$' + cipher.encrypt((s[i] + '$' + str(i)).encode("utf8")).hex()
    l = [i for i in range(0, len(s))]
    p = [i for i in range(0, len(s))]
    random.seed(k4)
    random.shuffle(p)
    for i in range(0, len(s)):
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(kl, AES.MODE_CFB, iv)
        l[p[i]] = iv.hex() + '$' + cipher.encrypt((str(tree.leaves[i].get_ind()) + '$' + str(i)).encode("utf8")).hex()
    return (d, c, l)



"""
Enc function initializer.

Jiangchen Zhu  <zjcsjtu@umich.edu>
Enhao Zhang    <ehzhang@umich.edu>
"""
from substr_enc.enc.encrypt import encrypt, key_gen
from substr_enc.enc.query import query_client



"""
Helper functions for suffix tree.

Jiangchen Zhu  <zjcsjtu@umich.edu>
"""

def get_path_helper(node):
    """Compute the path label of a node recursively."""
    if node.parent == None:
        # if this is a root node
        return ''
    else:
        return get_path_helper(node.parent) + node.edge_label


def get_initial_path_helper(node):
    """Compute the initial path label of a node recursively."""
    if node.parent == None:
        # if this is a root node
        return ''
    else:
        return node.parent.get_path() + node.edge_label[0]


def get_common_prefix(str_1, str_2):
    """Compute the longest common prefix of two strings."""
    longest_prefix = ''
    if not str_1 or not str_2:
        return longest_prefix
    shorter_len = min(len(str_1), len(str_2))
    for i in range(shorter_len):
        if str_1[i] == str_2[i]:
            longest_prefix += str_1[i]
        else:
            break
    return longest_prefix


def get_ind_helper(node, output):
    """Preorder traverse the tree and get the indices of the node."""
    if node == None:
        return
    if not node.children:
        output.append(node.leaf_label)
    for child in node.children:
        get_ind_helper(child, output)



"""
Substring-Searchable Encryption package setup.

Jiangchen Zhu  <zjcsjtu@umich.edu>
"""

from setuptools import setup

setup(
    name='substr_enc',
    version='0.1.4',
    packages=['substr_enc', 'substr_enc.enc', 'substr_enc.utils', 'substr_enc.conn'],
    include_package_data=True,
    install_requires=[
        'pytest==4.3.0',
        'pycryptodomex==3.8.1'
    ]
)



"""
Implementation of Query function in the paper.

Jiangchen Zhu  <zjcsjtu@umich.edu>
Enhao Zhang    <ehzhang@umich.edu>
Zhiqi Chen     <zhiqich@umich.edu>
"""

from substr_enc.utils import Node, SuffixTree
from substr_enc.model import LAMBDA
from substr_enc.enc import key_gen, encrypt
from Cryptodome.Cipher import AES
from Cryptodome import Random
from itertools import repeat
import hashlib
import random
import hmac

def query_client(k, p, d, c, l):
    """
    Query substring p in the ciphertext (d, c, l)

    k: the key list generated by key_gen
    p: the substring we are searching
    d, c, l: the ciphertext returned by function encrypt
    return: position of substring p in the original text
    """

    kd = k[0]
    kc = k[1]
    kl = k[2]
    k1 = k[3]
    k2 = k[4]
    k3 = k[5]
    k4 = k[6]
    m = len(p)
    t_list = []
    for i in range(m):
        prefix_p = p[:(i+1)]
        # f1
        h = hmac.new(key = k1, digestmod = hashlib.sha256)
        h.update(prefix_p.encode('utf-8'))
        f1 = h.digest()
        # f2
        h = hmac.new(key = k2, digestmod = hashlib.sha256)
        h.update(prefix_p.encode('utf-8'))
        f2 = h.digest()
        # T
        iv = Random.new().read(AES.block_size)
        obj = AES.new(f2, AES.MODE_CFB, iv)
        t_list.append(iv.hex() + "$" + obj.encrypt(f1).hex())
    h = hmac.new(key = k1, digestmod = hashlib.sha256)
    emptystring = ''
    h.update(emptystring.encode('utf-8'))
    root = h.digest()
    # client returns (root, t_list)
    parsed_list = d[root].split("$")
    for i in range(m):
        for j in range(128):
            ivenc = t_list[i].split("$")
            iv = bytes.fromhex(ivenc[0])
            enc = bytes.fromhex(ivenc[1])
            obj = AES.new(bytes.fromhex(parsed_list[j]), AES.MODE_CFB, iv)
            try:
                plaintext = obj.decrypt(enc)
                # f1 != perp
                parsed_list = d[plaintext].split("$")
                break
            except ValueError:
                print("Key incorrect or message corrupted")
            except KeyError:
                # print("wrong decryption")
                continue
    iv = bytes.fromhex(parsed_list[128])
    encW = bytes.fromhex(parsed_list[129])
    # server sends W (encW)
    obj = AES.new(kd, AES.MODE_CFB, iv)
    try:
        X = str(obj.decrypt(encW), "utf-8")
    except ValueError:
        return "No result: perp"
    # X != perp
    parsed_X = X.split("$")
    ind = int(parsed_X[0])
    leafpos = int(parsed_X[1])
    num = int(parsed_X[2])
    lenvar = int(parsed_X[3])
    f1 = bytes.fromhex(parsed_X[4])
    f_list = parsed_X[5:]
    h = hmac.new(key = k1, digestmod = hashlib.sha256)
    subp = p[:lenvar]
    h.update(subp.encode('utf-8'))
    if f1 != h.digest():
        return "No result: perp"
    for j in range(lenvar, m):
        for i in range(0, 128):
            ivencT = t_list[j].split("$")
            iv = bytes.fromhex(ivencT[0])
            encT = bytes.fromhex(ivencT[1])
            obj = AES.new(bytes.fromhex(f_list[i]), AES.MODE_CFB, iv)
            try:
                X = obj.decrypt(encT)
                # return "No result: perp"
            except:
                continue
    if ind == -1:
        return "no matching query found"
    m_seq = [i for i in range(0, m)]
    random.shuffle(m_seq)
    x_list = list(repeat(0, m))
    for i in range(m):
        random.seed(k3)
        prp_list = [i for i in range(0, len(c))]
        random.shuffle(prp_list)
        if ind + i >= len(c):
            return "no matching query found"
        x_list[m_seq[i]] = prp_list[ind + i]
    # client sends x_list, i.e. (x1, ..., xm)
    c_list = [i for i in range(0, m)]
    for i in range(m):
        c_list[i] = c[x_list[i]]
    # server sends c_list, i.e. (c1, ..., cm)
    for i in range(m):
        ivencC = c_list[m_seq[i]].split("$")
        iv = bytes.fromhex(ivencC[0])
        encC = bytes.fromhex(ivencC[1])
        obj = AES.new(kc, AES.MODE_CFB, iv)
        try:
            Y = str(obj.decrypt(encC), "utf-8")
        except ValueError:
            return "No result: perp"
        parsed_Y = Y.split("$")
        if int(parsed_Y[1]) != ind + i:
            return "No result: perp"
        if parsed_Y[0] != p[i]:
            return "no matching query found"
    num_seq = [i for i in range(0, num)]
    random.shuffle(num_seq)
    y_list = list(repeat(0, num))
    for i in range(num):
        random.seed(k4)
        prp_list = [i for i in range(0, len(c))]
        random.shuffle(prp_list)
        y_list[num_seq[i]] = prp_list[leafpos + i]
    # client sends y_list, i.e. (y1, ..., ynum)
    L_list = list(repeat(0, num))
    for i in range(num):
        L_list[i] = l[y_list[i]]
    # server sends L_list, i.e. (L1, ..., Lnum)
    A_list = []
    for i in range(num):
        ivencL = L_list[num_seq[i]].split("$")
        iv = bytes.fromhex(ivencL[0])
        encL = bytes.fromhex(ivencL[1])
        obj = AES.new(kl, AES.MODE_CFB, iv)
        try:
            dectext = str(obj.decrypt(encL), "utf-8")
        except ValueError:
            return "No result: perp"
        parsed_dec = dectext.split("$")
        if int(parsed_dec[1]) != leafpos + i:
            return "No result: perp"
        A_list.append(parsed_dec[0])
    A_list = list(map(int, A_list))
    A_list = sorted(A_list)
    return A_list
